from maedeep.dnn import (_forwardmapping)

def test_articulatory_to_contour(articulatory_parameters):
    contours_out = _forwardmapping.articulatory_to_contour(articulatory_parameters,
                                                           output_type="contour")
    assert len(contours_out) == 1
    assert len(contours_out[0].contours) == 4
    
def test_articulatory_to_area(articulatory_parameters):
    area_function_out = _forwardmapping.articulatory_to_area(articulatory_parameters,
                                                             output_type="area_function")
    assert hasattr(area_function_out, "area")
    assert hasattr(area_function_out, "length")
    assert area_function_out.length.shape == (40, 1)
    
def test_articulatory_to_task(articulatory_parameters):
    tasks_out = _forwardmapping.articulatory_to_task(articulatory_parameters)
    assert tasks_out.shape == (7, 1)
    
# def test_articulatory_to_tf(articulatory_parameters):
#     tf_out, freq = _forwardmapping.articulatory_to_transfer_function(articulatory_parameters)
#     assert tf_out.shape == (len(freq), 1)
    
def test_articulatory_to_formants(articulatory_parameters):
    formants_out = _forwardmapping.articulatory_to_formant(articulatory_parameters)
    assert formants_out.shape == (4, 1)

def contour_to_formants(contours):
    formants_out = _forwardmapping.contour_to_formant(contours)
    assert formants_out.shape == (4, 1)

def task_to_formants(tasks):
    formants_out = _forwardmapping.task_to_formant(tasks)
    assert formants_out.shape == (4, 1)
    
# def test_contours_to_area(contours, model, area_function):
#     area_function_out = _forwardmapping.contour_to_area(contours, model)
#     assert hasattr(area_function_out, "area")
#     assert hasattr(area_function_out, "length")
#     assert area_function_out.length.shape == (40, 1)
#     assert (area_function_out.area == area_function.area).all()
    
# def test_contour_to_task(contours, coeff, tasks):
#     tasks_out = _forwardmapping.contour_to_task(contours, coeff)
#     assert tasks_out.shape == (7, 1)
#     assert (tasks_out == tasks).all()
    
# def test_contour_to_transfer_function(contours, model, transfer_function):
#     tf_out, freq = _forwardmapping.contour_to_transfer_function(contours, model)
#     assert tf_out.shape == (len(freq), 1)
#     assert (tf_out == transfer_function[0]).all()
    
# def test_contour_to_formants(contours, model, formants):
#     formants_out = _forwardmapping.contour_to_formant(contours, model)
#     assert formants_out.shape == (4, 1)
#     assert (formants_out == formants).all()
    
# def test_area_to_transfer_function(area_function, transfer_function):
#     tf_out, freq = _forwardmapping.area_to_transfer_function(area_function)
#     assert tf_out.shape == (len(freq), 1)
#     assert (tf_out == transfer_function[0]).all()
    
# def test_area_to_formants(area_function, formants):
#     formants_out = _forwardmapping.area_to_formant(area_function)
#     assert formants_out.shape == (4, 1)
#     assert (formants_out == formants).all()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:04:55 2022

@author: benjamin
"""

import h5py
import numpy as np
from .dnntools import (
                        check_input_area,
                        dnn_joint_mapping,
                        dnn_mapping,
                        load_mapping_model,
                        normalization_model,
                        output_area,
                        output_contour,
                        reshape_input
                       )

def area_to_formant(areas, model_directory=None, model=None, norms=None):
    areas = check_input_area(areas)
    return dnn_mapping(areas, model_directory=model_directory,
                       input_space="area", output_space="acoustic",
                       model=model, norms=norms)

def articulatory_to_formant(articulatory_parameters, model_directory=None,
                            model=None, norms=None, joint=True):
    if joint:
        return dnn_joint_mapping(articulatory_parameters, 
                                 input_space="articulator",
                                 output_spaces=("acoustic", "task"),
                                 model=model, model_directory=model_directory)[0]
    else:
        return dnn_mapping(articulatory_parameters,
                           model_directory=model_directory,
                       input_space="articulator", output_space="acoustic",
                       model=model, norms=norms)

def articulatory_to_area(articulatory_parameters, model_directory=None,
                         model=None, norms=None, output_type="raw"):

    areas = dnn_mapping(articulatory_parameters, model_directory=model_directory,
                       input_space="articulator", output_space="area",
                       model=model, norms=norms)
    return output_area(areas, output_type)


def articulatory_to_task(articulatory_parameters, model_directory=None,
                         model=None, norms=None, joint=True):
    if joint:
        return dnn_joint_mapping(articulatory_parameters, 
                                 input_space="articulator",
                                 output_spaces=("acoustic", "task"),
                                 model=model, model_directory=model_directory)[1]
    else:
        return dnn_mapping(articulatory_parameters, model_directory=model_directory,
                           input_space="articulator", output_space="task",
                           model=model, norms=norms)

def articulatory_to_contour(articulatory_parameters, model_directory=None,
                            output_type="raw", model=None, norms=None):

    model, model_file = load_mapping_model(model_directory, "articulator",
                                     "contour")
    with h5py.File(model_file, "r") as hf:
        norms_input = hf["normalization_factors_input"][()]
        norms_output = hf["normalization_factors_output"][()]
        idx_var = hf["variance_index"][()]
        idx_invar = hf["invariance_index"][()]
        my_full = hf["full_mean"][()]
    nb_feat = model.get_config()["layers"][0]["config"]["batch_input_shape"][1]
    x = normalization_model(reshape_input(articulatory_parameters, nb_feat), norms_input, 0)
    predicted_contours = np.zeros((x.shape[0], my_full.shape[1]))
    predicted_contours[:, idx_var] = normalization_model(model.predict(x, verbose=0), norms_output, 1)
    predicted_contours[:, idx_invar] = np.repeat(my_full[:, idx_invar].reshape(1, -1),
                                                 x.shape[0], axis=0)

    return output_contour(predicted_contours, output_type)

def contour_to_area(contours, model_directory=None, output_type="raw",
                    model=None, norms=None):

    model, model_file = load_mapping_model(model_directory, "contour",
                                     "acoustic")
    with h5py.File(model_file, "r") as hf:
        norms_input = hf["normalization_factors_input"][()]
        if "normalization_factors_output" in hf.keys():
            norms_output = hf["normalization_factors_output"][()]
        else:
            norms_output = hf["normalization_factors_outpur"][()]
        idx_var = hf["variance_index"][()]

    x = np.array([contour.contours for contour in contours])
    x = normalization_model(x[:, idx_var], norms_input, 0)
    areas = normalization_model(model.predict(x, verbose=0), (norms_output, 0), 1)
    return output_area(areas, output_type)

def contour_to_formant(contours, model_directory=None, model=None, norms=None):

    model, model_file = load_mapping_model(model_directory, "contour",
                                     "acoustic")
    with h5py.File(model_file, "r") as hf:
        norms_input = hf["normalization_factors_input"][()]
        if "normalization_factors_output" in hf.keys():
            norms_output = hf["normalization_factors_output"][()]
        else:
            norms_output = hf["normalization_factors_outpur"][()]
        idx_var = hf["variance_index"][()]
    x = np.array([contour.contours for contour in contours])
    x = normalization_model(x[:, idx_var], norms_input, 0)
    return normalization_model(model.predict(x, verbose=0), norms_output, 1)

def contour_to_task(contours, model_directory=None, model=None, norms=None):

    model, model_file = load_mapping_model(model_directory, "contour",
                                     "task")
    with h5py.File(model_file, "r") as hf:
        norms_input = hf["normalization_factors_input"][()]
        if "normalization_factors_output" in hf.keys():
            norms_output = hf["normalization_factors_output"][()]
        else:
            norms_output = hf["normalization_factors_outpur"][()]
        idx_var = hf["variance_index"][()]
    x = np.array([contour.contours for contour in contours])
    x = normalization_model(x[:, idx_var], norms_input, 0)
    return normalization_model(model.predict(x, verbose=0), norms_output, 1)

def task_to_formant(task, model_directory=None, model=None, norms=None):
    return dnn_mapping(task, model_directory=model_directory,
                       input_space="task", output_space="acoustic",
                       model=model, norms=norms)













#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:04:55 2022

@author: benjamin
"""

import numpy as np
from maedeep._areafunction import (AreaFunction)
from maedeep._waveguide import Waveguide
from maedeep._contour import ( 
    extract_task_variables, 
    extract_task_variables_fast)
from maedeep.signaltools import get_impulse_response
from .modeltools import ( 
    check_model,
    contour_to_area_function, 
    vector_to_contours
    )

def area_to_formant(area_function):    
    wvg = Waveguide("area_function", area_function)
    wvg.computetransferfunction()
    return wvg.computeformants()

def area_to_transfer_function(area_function, df=50):
    wvg = Waveguide("area_function", area_function)
    return wvg.computetransferfunction(df=df)
  
def contour_to_area(contours, data, size_correction=None, min_area=1e-11):   
    if size_correction is None:
        size_correction = data["semi-polar coordinates"]["size correction"]
    alph, beta = [data["semi-polar coordinates"]["sagittal-area coefficients"][key]
                       for key in ["alpha", "beta"]]
    vp_map = data["semi-polar coordinates"]["map coeff"]
    iniva_tng = int(data["tongue"]["parameters"]["initial point"])
    
    c = size_correction * vp_map    
    afs = [contour_to_area_function(contour, c, iniva_tng, alph, 
                                    beta, min_area) 
              for contour in contours]
    
    AF = AreaFunction("area", np.array([x["area"] for x in afs]).T,
                      "length", np.array([x["length"] for x in afs]).T)

    AF.interpolate(40)
    return AF

def articulatory_to_area(articulatory_parameters, model=None, 
                         size_correction=None, 
                         min_area=1e-11):    
    contours = articulatory_to_contour(articulatory_parameters, model)
    data = check_model(model)
    area_function = contour_to_area(contours, data, size_correction, min_area)
    return area_function

def articulatory_to_contour(articulatory_parameters, model=None):
    return vector_to_contours(articulatory_parameters, model)

def articulatory_to_formant(articulatory_parameters, model=None, 
                            size_correction=None, min_area=1e-11):
    return area_to_formant(articulatory_to_area(articulatory_parameters, 
                                                model=model,
                                                size_correction=size_correction,
                                                min_area=min_area))    

def articulatory_to_impulse_response(articulatory_parameters, 
                                     model=None, df=50, min_area=1e-11):
    transfer_function = articulatory_to_transfer_function(articulatory_parameters,
                                                          model=model, df=df,
                                                          min_area=min_area)[0]
    return get_impulse_response(transfer_function)

def articulatory_to_task(articulatory_parameters, model=None, verbosity=False,
                         method="fast"):
    data = check_model(model)    
    size_correction = data["semi-polar coordinates"]["size correction"]    
    vp_map = data["semi-polar coordinates"]["map coeff"]
    c = size_correction * vp_map    
    if method == "fast":
        contours = articulatory_to_contour(articulatory_parameters, model)
        return extract_task_variables_fast(contours, c)
    else:
        return contour_to_task(articulatory_to_contour(articulatory_parameters, 
                                                       model), 
                            c, verbosity=verbosity)

def articulatory_to_transfer_function(articulatory_parameters, model=None, 
                                      df=50, min_area=1e-11):
    return area_to_transfer_function(articulatory_to_area(articulatory_parameters, 
                                                    model,
                                                    min_area=min_area), df=df)

def contour_to_formant(contours, model=None, size_correction=None):
    data = check_model(model)
    return area_to_formant(contour_to_area(contours, data, size_correction))
    
def contour_to_task(contours, c, verbosity=False):
    if verbosity:
        from tqdm import tqdm
        return np.array([extract_task_variables(ctr, c) for ctr in tqdm(contours)]).T
    else:
        return np.array([extract_task_variables(ctr, c) for ctr in contours]).T

def contour_to_transfer_function(contours, model=None, df=50,
                                 size_correction=None):
    data = check_model(model)
    return area_to_transfer_function(contour_to_area(contours, data, 
                                                     size_correction), df=df)

def transfer_function_to_formant(transfer_function, freq):
    wvg = Waveguide("transfer_function", transfer_function,
                    "freq", freq)
    return wvg.computeformants()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:04:55 2022

@author: benjamin
"""

import numpy as np
from maedeep.signaltools import (
    vtl_estimation
    )
from .modeltools import (
    check_model
    )
from .optimtools import Optimizer
import maedeep.dnn as dnn

def acoustic_to_articulatory(formants, weight=0, vtln=True, model=None,
                         init_estimate='model', verbosity=1, fixed_size=None,
                         min_area=0, threshold=0, art2ac_model=None):

    if vtln:
        length = vtl_estimation(formants)
    else: 
        length = 0.16273747
    if isinstance(init_estimate, str):
        if init_estimate == 'model':
            articulatory_params = dnn.acoustic_to_articulatory(formants)
        elif init_estimate == 'random':
            articulatory_params = np.random.uniform(-3, 3, (7, formants.shape[0]))
    else:
        articulatory_params = np.array(init_estimate)
    data = check_model(model)
    c = data["semi-polar coordinates"]["size correction"]
    x = articulatory_params.flatten().tolist()
    optim = Optimizer("vocal_model", data, "weight", weight,
                                 'verbosity', verbosity,
                                 'min_area', min_area,
                                 'threshold', threshold,
                                 'acoustic_model', art2ac_model)

    if fixed_size is not None:
        if fixed_size == 0:
            fixed_size = c / length * 0.16273747
        optim.fixed_size_factor = fixed_size
    else:
        fixed_size = c / length * 0.16273747
        x += [c / length * 0.16273747]
        optim.fixed_size_factor = None
    optim.global_optimization(formants, x0=x)
    return optim








# -*- coding: utf-8 -*-
"""
Created on Tue May 30 10:52:13 2023

@author: belie
"""

from ._forwardmapping import articulatory_to_formant
import copy
import numpy as np
import os
from scipy.optimize import minimize
from sklearn.metrics import mean_squared_error
import tensorflow as tf


class Optimizer:
    name = None
    vocal_model = None
    weight = None
    threshold = 0
    nb_attempts = 3
    verbosity = True
    parallel = 1
    repetitions = 1
    maxiter = 200
    solution = None
    fixed_size_factor = None
    optimal_cost = None
    min_area = 0
    acoustic_model = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

    def copy(self):
        return copy.deepcopy(self)

    def global_optimization(self, refs, x0=None):
        """ Performs global optimization """

        if x0 is not None:
            global_min_cost = objective(x0, self, refs)
            global_min_x = [x for x in x0]
        else:
            global_min_cost = np.inf
            global_min_x = None

        if self.parallel <= 1 or self.repetitions <= 1:
            for k in range(self.repetitions):
                local_min_cost, local_min_x = self.repetition(x0,
                                                         global_min_cost, k,
                                                         refs)

                global_checks = check_global_optimization(local_min_cost,
                                                          local_min_x,
                                                          global_min_cost, 
                                                          global_min_x)
                global_min_cost, global_min_x = global_checks
        else:
            from joblib import Parallel, delayed
            if os.name == "nt":
                backend = "threading"
            else:
                backend = "loky"
            solutions = Parallel(n_jobs=self.parallel, backend=backend,
                             prefer="processes")(delayed(self.repetition)(x0,
                                                                      None, k,
                                                                      refs) for
                                                     k in range(self.repetitions))
            for solution in solutions:
                local_min_cost, local_min_x = solution
                global_checks = check_global_optimization(local_min_cost,
                                                          local_min_x,
                                              global_min_cost, global_min_x)
                global_min_cost, global_min_x = global_checks

        if self.verbosity:
            print("Optimization completed", flush=True)
            print("Global minimum: ", global_min_cost, flush=True)

        self.solution = global_min_x
        self.optimal_cost = global_min_cost
        return global_min_x

    def local_optimization(self, x0, refs):
        options = {'maxiter': self.maxiter * len(x0)}
        if self.fixed_size_factor is None:
            bounds = ((-3, 3), ) * (len(x0) - 1) + ((0, 2), )
        else:
            bounds = ((-3, 3), ) * len(x0)
        args = (self, refs)
        result = minimize(objective, x0, args=args, method="nelder-mead",
                              options=options, bounds=bounds)
        return result["x"], result["fun"]

    def repetition(self, x0_start, global_min_cost, k, refs):

        """ Perform one otpimization run """
        nb_f = refs.shape[1]
        if x0_start is not None:
            min_x = [xx for xx in x0_start]
        else:
            min_x = (6*np.random.rand(7*nb_f)-3).tolist() # + [1]
            if self.fixed_size_factor is None:
                min_x += [1]

        rem_attempts = self.nb_attempts
        local_min_cost = objective(min_x, self, refs)
        isContinue = True

        while isContinue:
            new_x, new_cost = self.local_optimization(min_x, refs=refs)
            (min_x, rem_attempts,
             local_min_cost, isContinue) = check_local_optimization(new_x,
                                            new_cost, local_min_cost,
                                         self.nb_attempts, min_x,
                                         self.threshold, rem_attempts)

            if self.verbosity >= 1 and global_min_cost is not None:
                print("Current repetition: ", k+1, " /",
                      self.repetitions, flush=True)
                print("Global cost function ", global_min_cost, flush=True)
                print("Local cost function: ", local_min_cost, flush=True)
                print("New cost :", new_cost, flush=True)
                print("Number of remaining attempts :",
                      rem_attempts, flush=True)

        return local_min_cost, min_x

def check_global_optimization(local_min_cost, local_min_x,
                              global_min_cost, global_min_x):
    """ Check whether the local minimum in the local loop
    is temporary a global one """
    if local_min_cost < global_min_cost:
        global_min_x = local_min_x
        global_min_cost = local_min_cost

    return global_min_cost, global_min_x

def check_local_optimization(new_x, new_cost, local_min_cost,
                             nb_attempts, min_x, threshold, rem_attempts):
    """ Check whether the local minimum is temporary a global one """

    isContinue = True
    gain = np.abs((new_cost - local_min_cost) / local_min_cost)
    if new_cost < local_min_cost:
        local_min_cost = new_cost
        min_x = new_x

        if gain > threshold:
            rem_attempts = nb_attempts
        else:
            rem_attempts -= 1
            if rem_attempts == 0:
                isContinue = False
    else:
        rem_attempts -= 1
        if rem_attempts == 0:
            isContinue = False
    return min_x, rem_attempts, local_min_cost, isContinue

def formant_cost(ref, est):
    return mean_squared_error(ref.flatten()/ref.flatten(), 
                              est.flatten()/ref.flatten(), squared=False)

def get_formants_model(x, acoustic_model):
    x_scaled = acoustic_model[1][0].transform(np.array(x).reshape(1,-1))
    x_scaled = tf.convert_to_tensor(x_scaled, dtype=tf.float32)
    f = acoustic_model[0](x_scaled).numpy()    
    return acoustic_model[1][1].inverse_transform(f)


def reshape_solution(x, size_correction=True):
    if size_correction:
        return np.reshape(x[:-1], (7, -1))
    else:
        return np.reshape(x, (7, -1))

def objective(x, *args):
    optim, refs = args
    return solution_to_cost(x, optim, refs)

def solution_to_cost(x, optim, formants_ref):
    if optim.fixed_size_factor is None:
        p = reshape_solution(x, size_correction=True)
        c = x[-1]
    else:
        p = reshape_solution(x, size_correction=False)
        c = optim.fixed_size_factor
    formants_estimated = np.nan_to_num(solution_2_formants(p, c,
                                                       optim.vocal_model,
                                                       optim.min_area, 
                                                       optim.acoustic_model),
                                       0)
    acoustic_cost = formant_cost(formants_ref, formants_estimated)
    if optim.weight != 0:
        var_p = solution_2_var(p)
        total_cost = acoustic_cost + optim.weight * var_p
    else:
        total_cost =  acoustic_cost
    if optim.verbosity == 2:
        print("Cost function is", total_cost, flush=True)
    return total_cost

def solution_2_formants(p, sc, model, min_area=0, ac_model=None):
    if ac_model is None:
        return articulatory_to_formant(p, model, 
                                       size_correction=sc, min_area=min_area)
    else:
        return get_formants_model(p, ac_model)

def solution_2_var(p):
    return np.linalg.norm(np.std(p, axis=1))**2
